from greyhorse.app.abc.operators import AssignOperator
from greyhorse.app.abc.providers import SharedProvider
from greyhorse.app.boxes import SharedRefBox
from greyhorse.app.entities.controllers import SyncController, operator
from greyhorse.app.entities.services import SyncService, provider
from greyhorse.maybe import Maybe, Nothing

from .common.functional import FunctionalOperator


class FunctionalOperatorService(SyncService):
    def __init__(self) -> None:
        super().__init__()
        self._res = Nothing

    def get_res(self) -> Maybe[FunctionalOperator]:
        return self._res

    def set_res(self, res: Maybe[FunctionalOperator]) -> None:
        self._res = res

    @provider(SharedProvider[FunctionalOperator])
    def create_op(self):
        if not self._res:
            return None
        return SharedRefBox(self.get_res, lambda v: v)


class FunctionalOperatorCtrl(SyncController):
    def __init__(self, svc: FunctionalOperatorService) -> None:
        super().__init__()
        self._svc = svc

    @operator(FunctionalOperator)
    def create_op(self):
        return AssignOperator[FunctionalOperator](self._svc.get_res, self._svc.set_res)



from abc import ABC, abstractmethod
from collections.abc import Callable

from greyhorse.maybe import Maybe
from greyhorse.utils.types import TypeWrapper


class Selector[K, T](TypeWrapper[K, T], ABC):
    @abstractmethod
    def has(self, key: K) -> bool: ...

    @abstractmethod
    def get(self, key: K) -> Maybe[T]: ...


class ListSelector[K, T](Selector[K, T], ABC):
    @abstractmethod
    def items(self, filter_fn: Callable[[K], bool] | None = None) -> list[tuple[K, T]]: ...



import asyncio
from abc import ABC
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, Self, TypeVar

from pydantic.main import BaseModel as PydanticModel

from .base import AbstractModel, IdType
from .fields import ModelFieldsMixin

if TYPE_CHECKING:
    from ..repositories.base import ModelRepository


CreateSchemaType = TypeVar('CreateSchemaType', bound=PydanticModel)
UpdateSchemaType = TypeVar('UpdateSchemaType', bound=PydanticModel)


class Model(AbstractModel[IdType], ModelFieldsMixin, ABC):
    _repo: 'ModelRepository[IdType, Self]'

    @classmethod
    def bind(cls, repository: 'ModelRepository[IdType, Self]') -> None:
        cls._repo = repository

    @classmethod
    async def construct(cls, data: Mapping[str, Any], **kwargs) -> Self:
        return await cls._repo.construct(data, **kwargs)

    @classmethod
    async def construct_all(
        cls, objects: Sequence[Mapping[str, Any] | None], **kwargs
    ) -> Sequence[Self | None]:
        loop = asyncio.get_event_loop()
        awaitables = list()

        for data in objects:
            if data is None:
                future = loop.create_future()
                future.set_result(None)
                awaitables.append(future)
            else:
                awaitables.append(cls.construct(data, **kwargs))

        return await asyncio.gather(*awaitables)

    @classmethod
    def prepare_for_create(cls, data: Mapping[str, Any]) -> Mapping[str, Any]:
        return data

    @classmethod
    def prepare_for_update(cls, data: Mapping[str, Any]) -> Mapping[str, Any]:
        return data

    @classmethod
    def prepare_for_delete(cls, obj: Self) -> Self:
        return obj

    #
    # Retrieve operations
    #

    @classmethod
    async def get(cls, id_value: IdType, **kwargs) -> Self | None:
        return await cls._repo.get(id_value, **kwargs)

    @classmethod
    async def get_any(cls, indices: Sequence[IdType], **kwargs) -> Sequence[Self | None]:
        return await cls._repo.get_any(indices, **kwargs)

    @classmethod
    async def exists(cls, id_value: IdType, **kwargs) -> bool:
        return await cls._repo.exists(id_value, **kwargs)

    async def load(self, only: Sequence[str] | None = None) -> bool:
        return await self._repo.load(self, only)

    #
    # Modification operations
    #

    @classmethod
    async def create(cls, data: CreateSchemaType | Mapping[str, Any], **kwargs) -> Self | None:
        create_data = data if isinstance(data, Mapping) else data.dict(exclude_unset=True)

        if create_data:
            create_data = cls.prepare_for_create(create_data)

        return await cls._repo.create(create_data, **kwargs)

    @classmethod
    async def get_or_create(
        cls, id_value: Any, data: CreateSchemaType | Mapping[str, Any], **kwargs
    ) -> tuple[Self | None, bool]:
        if instance := await cls.get(id_value, **kwargs):
            return instance, False
        return await cls.create(data, **kwargs), True

    async def update(self, data: UpdateSchemaType | Mapping[str, Any], **kwargs) -> bool:
        update_data = data if isinstance(data, Mapping) else data.dict(exclude_unset=True)

        if update_data:
            update_data = self.prepare_for_update(update_data)

        return await self._repo.update_by_id(self.get_id_value(), update_data, **kwargs)

    @classmethod
    async def update_by_id(
        cls, id_value: IdType, data: UpdateSchemaType | Mapping[str, Any], **kwargs
    ) -> bool:
        update_data = data if isinstance(data, Mapping) else data.dict(exclude_unset=True)

        if update_data:
            update_data = cls.prepare_for_update(update_data)

        return await cls._repo.update_by_id(id_value, update_data, **kwargs)

    async def save(self, **kwargs) -> bool:
        return await self._repo.save(self, **kwargs)

    @classmethod
    async def save_all(cls, objects: Sequence[Self], **kwargs) -> bool:
        return await cls._repo.save_all(objects, **kwargs)

    async def delete(self) -> bool:
        return await self._repo.delete(self)

    @classmethod
    async def delete_all(cls, indices: Sequence[IdType] | None = None) -> int:
        return await cls._repo.delete_all(indices)

    @classmethod
    async def delete_by_id(cls, id_value: IdType) -> bool:
        return await cls._repo.delete_by_id(id_value)



from greyhorse.app.schemas.components import ComponentConf, ModuleConf, ProvidersConf
from greyhorse.app.schemas.elements import CtrlConf, SvcConf

from ..common.functional import FunctionalOperator, FunctionalOpProvider
from ..common.resources import (
    DictCtxProvider,
    DictMutCtxProvider,
    DictResContext,
    MutDictResContext,
)
from ..components.functional import DictOperatorCtrl, DictOperatorService
from ..components.resources import DictProviderService


def __init__():
    return ModuleConf(
        enabled=True,
        can_provide=[FunctionalOperator],
        components={
            'domain': ComponentConf(
                enabled=True,
                provider_imports=[
                    ProvidersConf(resource=DictResContext, providers=[DictCtxProvider]),
                    ProvidersConf(resource=MutDictResContext, providers=[DictMutCtxProvider]),
                ],
                services=[SvcConf(type=DictProviderService)],
            ),
            'app': ComponentConf(
                enabled=True,
                provider_grants=[
                    ProvidersConf(resource=DictResContext, providers=[DictCtxProvider]),
                    ProvidersConf(resource=MutDictResContext, providers=[DictMutCtxProvider]),
                ],
                provider_imports=[
                    ProvidersConf(resource=FunctionalOperator, providers=[FunctionalOpProvider])
                ],
                services=[SvcConf(type=DictOperatorService, resources=[DictResContext])],
                controllers=[CtrlConf(type=DictOperatorCtrl)],
            ),
        },
    )



from abc import ABC
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .controllers import Controller
    from .services import Service


class Visitor(ABC):
    def visit_controller(self, controller: 'Controller') -> None: ...

    def visit_service(self, service: 'Service') -> None: ...

    def start_component(self, component) -> None: ...

    def finish_component(self, component) -> None: ...

    def start_module(self, module) -> None: ...

    def finish_module(self, module) -> None: ...



from copy import deepcopy
from functools import partial

from greyhorse.app.abc.providers import BorrowError, BorrowMutError, ForwardError
from greyhorse.app.boxes import (
    ForwardBox,
    MutCtxRefBox,
    MutRefBox,
    OwnerCtxRefBox,
    OwnerRefBox,
    SharedCtxRefBox,
    SharedRefBox,
)
from greyhorse.app.contexts import (
    MutCtxCallbacks,
    SyncContext,
    SyncMutContext,
    SyncMutContextWithCallbacks,
)
from greyhorse.maybe import Just


def test_shared() -> None:
    instance = SharedRefBox[int](lambda: 123)

    res = instance.borrow()
    assert res.is_ok()

    res = instance.borrow()
    assert res.is_ok()

    unwrapped = res.unwrap()
    assert unwrapped == 123
    instance.reclaim(unwrapped)


def test_mut() -> None:
    instance = MutRefBox[int](lambda: 123)

    res = instance.acquire()
    assert res.is_ok()
    unwrapped = res.unwrap()

    res = instance.acquire()
    assert res.is_err()
    assert res.unwrap_err() == BorrowMutError.AlreadyBorrowed(name='int')

    instance.release(unwrapped)

    res = instance.acquire()
    assert res.is_ok()

    assert res.unwrap() == 123


def test_owner() -> None:
    instance = OwnerRefBox[int, str](lambda: 123, lambda: '123')

    res = instance.borrow()
    assert res.is_ok()

    res = instance.borrow()
    assert res.is_ok()

    unwrapped = res.unwrap()
    assert unwrapped == 123

    res = instance.acquire()
    assert res.is_err()
    assert res.unwrap_err() == BorrowMutError.BorrowedAsImmutable(name='str')

    instance.reclaim(unwrapped)
    instance.reclaim(unwrapped)

    res = instance.acquire()
    assert res.is_ok()

    unwrapped = res.unwrap()
    assert unwrapped == '123'

    res = instance.borrow()
    assert res.is_err()
    assert res.unwrap_err() == BorrowError.BorrowedAsMutable(name='int')

    instance.release(unwrapped)

    res = instance.borrow()
    assert res.is_ok()

    unwrapped = res.unwrap()
    assert unwrapped == 123


def test_shared_context() -> None:
    instance = SharedCtxRefBox[int](SyncContext, lambda: 123)

    res = instance.borrow()
    assert res.is_ok()

    res = instance.borrow()
    assert res.is_ok()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncContext)

    with ctx as data:
        assert data == 123


def test_mut_context() -> None:
    instance = MutCtxRefBox[int](SyncMutContext, lambda: 123)

    res = instance.acquire()
    assert res.is_ok()
    unwrapped = res.unwrap()

    res = instance.acquire()
    assert res.is_err()
    assert res.unwrap_err() == BorrowMutError.AlreadyBorrowed(name='int')

    instance.release(unwrapped)

    res = instance.acquire()
    assert res.is_ok()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncMutContext)

    with ctx as data:
        assert data == 123


def test_owner_context() -> None:
    instance = OwnerCtxRefBox[int, str](SyncContext, SyncMutContext, lambda: 123, lambda: '123')

    res = instance.borrow()
    assert res.is_ok()

    res = instance.borrow()
    assert res.is_ok()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncContext)

    with ctx as data:
        assert data == 123

    res = instance.acquire()
    assert res.is_err()
    assert res.unwrap_err() == BorrowMutError.BorrowedAsImmutable(name='str')

    instance.reclaim(ctx)
    instance.reclaim(ctx)

    res = instance.acquire()
    assert res.is_ok()

    mut_ctx = res.unwrap()
    assert isinstance(mut_ctx, SyncMutContext)

    with mut_ctx as data:
        assert data == '123'

    res = instance.borrow()
    assert res.is_err()
    assert res.unwrap_err() == BorrowError.BorrowedAsMutable(name='int')

    instance.release(mut_ctx)

    res = instance.borrow()
    assert res.is_ok()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncContext)

    with ctx as data:
        assert data == 123


def test_owning_context_read_write() -> None:
    value = {'counter': 1}

    instance = OwnerCtxRefBox[dict, dict](
        SyncContext,
        SyncMutContextWithCallbacks,
        partial(deepcopy, value),
        partial(deepcopy, value),
        mut_params=dict(callbacks=MutCtxCallbacks(on_apply=Just(lambda v: value.update(v)))),
    )

    res = instance.borrow()
    assert res.is_ok()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncContext)

    with ctx as data:
        assert data == {'counter': 1}

    instance.reclaim(ctx)

    res = instance.acquire()
    assert res.is_ok()

    mut_ctx = res.unwrap()
    assert isinstance(mut_ctx, SyncMutContext)

    with mut_ctx as data:
        assert data == {'counter': 1}
        data['counter'] += 1

    instance.release(mut_ctx)

    res = instance.borrow()
    assert res.is_ok()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncContext)

    with ctx as data:
        assert data == {'counter': 1}

    instance.reclaim(ctx)

    res = instance.acquire()
    assert res.is_ok()

    mut_ctx = res.unwrap()
    assert isinstance(mut_ctx, SyncMutContext)

    with mut_ctx as data:
        assert data == {'counter': 1}
        data['counter'] += 1
        mut_ctx.apply()

    instance.release(mut_ctx)

    res = instance.borrow()
    assert res.is_ok()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncContext)

    with ctx as data:
        assert data == {'counter': 2}


def test_forward() -> None:
    instance = ForwardBox[int]()

    assert not instance

    res = instance.take()
    assert res.is_err()
    assert res.unwrap_err() == ForwardError.Empty(name='int')

    instance.accept(123)

    res = instance.take()
    assert res.is_ok()

    unwrapped = res.unwrap()
    assert unwrapped == 123

    res = instance.take()
    assert res.is_err()
    assert res.unwrap_err() == ForwardError.Empty(name='int')

    instance.drop(unwrapped)



import asyncio
import threading
from typing import get_type_hints, override

from greyhorse.result import Ok, Result

from ..abc.providers import Provider
from ..abc.services import ProviderMember, Service, ServiceError, ServiceState, ServiceWaiter


def provider(provider_type: type[Provider]):
    def decorator(func: classmethod):
        class_name, method_name = func.__qualname__.split('.')
        hints = get_type_hints(func, include_extras=True)
        ret_type = hints.pop('return', None)
        func.__provider__ = ProviderMember(
            class_name,
            method_name,
            provider_type.__wrapped_type__,
            provider_type,
            method=func,
            params=hints,
            ret_type=ret_type,
        )
        return func

    return decorator


class SyncService(Service):
    def __init__(self) -> None:
        super().__init__()
        self._provided_providers: dict[type[Provider], Provider] = {}
        self._waiter = threading.Event()

    @override
    @property
    def waiter(self) -> ServiceWaiter:
        return ServiceWaiter.Sync(self._waiter)

    @override
    def setup(self, *args, **kwargs) -> Result[ServiceState, ServiceError]:
        if self.state == ServiceState.Active:
            return Ok(self.state)

        self._switch_to_active()
        return Ok(self.state)

    @override
    def teardown(self, *args, **kwargs) -> Result[ServiceState, ServiceError]:
        if self.state == ServiceState.Idle:
            return Ok(self.state)

        self._switch_to_idle()
        return Ok(self.state)

    @override
    def _switch_to_idle(self) -> None:
        self._state = ServiceState.Idle

    @override
    def _switch_to_active(self, started: bool = False) -> None:
        if started:
            self._waiter.clear()
        else:
            self._waiter.set()

        self._state = ServiceState.Active(started=not self._waiter.is_set())


class AsyncService(Service):
    def __init__(self) -> None:
        super().__init__()
        self._provided_providers: dict[type[Provider], Provider] = {}
        self._waiter = asyncio.Event()

    @override
    @property
    def waiter(self) -> ServiceWaiter:
        return ServiceWaiter.Async(self._waiter)

    @override
    async def setup(self, *args, **kwargs) -> Result[ServiceState, ServiceError]:
        if self.state == ServiceState.Active:
            return Ok(self.state)

        await self._switch_to_active()
        return Ok(self.state)

    @override
    async def teardown(self, *args, **kwargs) -> Result[ServiceState, ServiceError]:
        if self.state == ServiceState.Idle:
            return Ok(self.state)

        await self._switch_to_idle()
        return Ok(self.state)

    @override
    async def _switch_to_idle(self) -> None:
        self._state = ServiceState.Idle

    @override
    async def _switch_to_active(self, started: bool = False) -> None:
        if started:
            self._waiter.clear()
        else:
            self._waiter.set()

        self._state = ServiceState.Active(started=not self._waiter.is_set())



from mypy.mro import calculate_mro
from mypy.nodes import GDEF, Block, ClassDef, SymbolTable, SymbolTableNode, TypeInfo
from mypy.plugin import AttributeContext, ClassDefContext, DynamicClassDefContext, Plugin
from mypy.types import Instance, Type


def _base_cls_hook(ctx: ClassDefContext) -> None:
    if ctx.cls.fullname == 'greyhorse.error.ErrorCase':
        pass


def _dynamic_class_hook(ctx: DynamicClassDefContext) -> None:
    cls = ClassDef(ctx.name, Block([]))
    cls.fullname = ctx.api.qualified_name(ctx.name)
    sym = ctx.api.lookup_fully_qualified_or_none(ctx.api.type.fullname)
    base = Instance(sym.node, [])

    info = TypeInfo(SymbolTable(), cls, ctx.api.cur_mod_id)
    info.bases = [base]
    cls.info = info
    calculate_mro(info)
    ctx.api.add_symbol_table_node(ctx.name, SymbolTableNode(GDEF, info))


def _class_attribute_hook(ctx: AttributeContext) -> Type:
    return ctx.type


class APIClientPlugin(Plugin):
    def get_base_class_hook(self, fullname: str):
        sym = self.lookup_fully_qualified(fullname)
        if sym and isinstance(sym.node, TypeInfo) and fullname.startswith('greyhorse'):
            print(f'base_class: {fullname}')
            return _base_cls_hook
        return None

    def get_dynamic_class_hook(self, fullname: str):
        sym = self.lookup_fully_qualified(fullname)
        if sym and isinstance(sym.node, TypeInfo):
            for name in ('greyhorse.enum', 'greyhorse.result'):
                if fullname.startswith(name):
                    print(f'dynamic_class: {fullname}')
                    return _dynamic_class_hook
        return None

    # def get_class_attribute_hook(self, fullname: str) -> Callable[[AttributeContext], Type] | None:
    #     sym = self.lookup_fully_qualified(fullname)
    #     if sym and isinstance(sym.node, TypeInfo):
    #         if fullname.startswith('greyhorse'):
    #             print(f'class_attribute: {fullname}')
    #             return _class_attribute_hook


def plugin(version: str):
    return APIClientPlugin



import decimal
from collections.abc import Callable
from typing import Any

from orjson import orjson


def stringify(obj):
    if isinstance(obj, decimal.Decimal):
        return str(obj)
    raise TypeError


def dumps_raw(
    content: Any,
    use_indent: bool = False,
    sort_keys: bool = False,
    stringify_fn: Callable[[Any], str] | None = None,
    **kwargs,
) -> bytes:
    option = orjson.OPT_SERIALIZE_UUID
    if use_indent:
        option |= orjson.OPT_INDENT_2
    if sort_keys:
        option |= orjson.OPT_SORT_KEYS
    return orjson.dumps(content, option=option, default=stringify_fn or stringify, **kwargs)


def dumps(
    content: Any,
    use_indent: bool = False,
    sort_keys: bool = False,
    stringify_fn: Callable[[Any], str] | None = None,
    **kwargs,
) -> str:
    return dumps_raw(content, use_indent, sort_keys, stringify_fn, **kwargs).decode('utf-8')


def loads(data: bytes | bytearray | memoryview | str) -> Any:
    return orjson.loads(data)



from abc import ABC, abstractmethod
from collections.abc import Awaitable, Mapping
from typing import NewType, override

from greyhorse.app.abc.providers import Provider
from greyhorse.app.registries import MutDictRegistry
from greyhorse.maybe import Just, Maybe, Nothing


class ProviderRegistry:
    __slots__ = ('_storage',)

    def __init__(self) -> None:
        self._storage = MutDictRegistry[tuple[type, str], Provider]()

    def add(self, conn_type: type, name: str, provider: Provider) -> bool:
        return self._storage.add((conn_type, name), provider)

    def remove(self, conn_type: type, name: str) -> bool:
        return self._storage.remove((conn_type, name))

    def has(self, conn_type: type, name: str | None = None) -> bool:
        items = self._storage.items(
            lambda key: key[0] == conn_type and (key[1] == name if name else True)
        )
        return len(items) > 0

    def get[T: Provider](self, conn_type: type, name: str | None = None) -> Maybe[T]:
        items = self._storage.items(
            lambda key: key[0] == conn_type and (key[1] == name if name else True)
        )
        if items:
            return Just(items[0][1])
        return Nothing

    def __len__(self) -> int:
        return len(self._storage)


ConnectionProviderRegistry = NewType('ConnectionProviderRegistry', ProviderRegistry)
SessionProviderRegistry = NewType('SessionProviderRegistry', ProviderRegistry)


class DataStorageEngine(ABC):
    def __init__(self, name: str) -> None:
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    @abstractmethod
    def active(self) -> bool: ...

    @abstractmethod
    def get_provider[P: Provider](self, prov_type: type[P]) -> Maybe[P]: ...

    def start(self) -> Awaitable[None] | None:
        pass

    def stop(self) -> Awaitable[None] | None:
        pass


class DataStorageEngineFactory(ABC):
    @abstractmethod
    def create_engine(self, name: str, *args, **kwargs) -> DataStorageEngine: ...

    @abstractmethod
    def destroy_engine(self, name: str) -> bool: ...

    @abstractmethod
    def get_engine_names(self) -> list[str]: ...

    @abstractmethod
    def get_engine(self, name: str) -> Maybe[DataStorageEngine]: ...

    @abstractmethod
    def get_engines(
        self, names: list[str] | None = None
    ) -> Mapping[str, DataStorageEngine]: ...


class SimpleDataStorageFactory[T: DataStorageEngine](DataStorageEngineFactory, ABC):
    def __init__(self) -> None:
        self._engines = MutDictRegistry[str, DataStorageEngine]()

    @override
    def destroy_engine(self, name: str) -> bool:
        return self._engines.remove(name)

    @override
    def get_engine_names(self) -> list[str]:
        return [k for k, _ in self._engines.items()]

    @override
    def get_engine(self, name: str) -> Maybe[T]:
        return self._engines.get(name)

    @override
    def get_engines(self, names: list[str] | None = None) -> Mapping[str, T]:
        engines = {}

        if not names:
            names = self.get_engine_names()

        for name in names:
            if engine := self.get_engine(name).unwrap_or_none():
                engines[name] = engine

        return engines



from typing import Any

from greyhorse.app.abc.operators import Operator
from greyhorse.app.abc.providers import Provider
from greyhorse.app.abc.visitor import Visitor
from greyhorse.app.entities.components import Component
from greyhorse.app.private.res_manager import ResourceManager
from greyhorse.app.registries import MutDictRegistry
from greyhorse.app.schemas.components import ModuleConf
from greyhorse.error import Error, ErrorCase
from greyhorse.logging import logger
from greyhorse.result import Ok, Result


class ModuleError(Error):
    namespace = 'greyhorse.app.module'

    Component = ErrorCase(
        msg='{path}: Component error in module, details: "{details}"', path=str, details=str
    )

    Resource = ErrorCase(
        msg='{path}: Resource error in module: "{details}"', path=str, details=str
    )


class Module:
    def __init__(self, path: str, conf: ModuleConf, components: list[Component]) -> None:
        self._conf = conf
        self._path = path
        self._rm = ResourceManager()

        self._operators: list[Operator] = []

        self._resources = MutDictRegistry[type, Any]()
        self._providers = MutDictRegistry[type[Provider], Provider]()
        self._components: dict[str, Component] = {c.name: c for c in components}

    @property
    def path(self) -> str:
        return self._path

    def accept_visitor(self, visitor: Visitor) -> None:
        visitor.start_module(self)
        for comp in self._components.values():
            comp.accept_visitor(visitor)
        visitor.finish_module(self)

    def add_provider[T](self, prov_type: type[Provider[T]], provider: Provider[T]) -> bool:
        for prov_claim in self._conf.provider_claims:
            if prov_type in prov_claim.providers:
                return self._providers.add(prov_type, provider)
        return False

    def remove_provider[T](self, prov_type: type[Provider[T]]) -> bool:
        for prov_claim in self._conf.provider_claims:
            if prov_type in prov_claim.providers:
                return self._providers.remove(prov_type)
        return False

    def add_resource(self, res_type: type, resource: Any) -> bool:
        if res_type in self._conf.resource_claims:
            return self._resources.add(res_type, resource)
        return False

    def remove_resource(self, res_type: type) -> bool:
        if res_type in self._conf.resource_claims:
            return self._resources.remove(res_type)
        return False

    def add_operator[T](self, operator: Operator[T]) -> bool:
        for res_type in self._conf.can_provide:
            if issubclass(res_type, operator.wrapped_type):
                self._operators.append(operator)
                return True
        return False

    def remove_operator[T](self, operator: Operator[T]) -> bool:
        for res_type in self._conf.can_provide:
            if issubclass(res_type, operator.wrapped_type):
                self._operators.remove(operator)
                return True
        return False

    def setup(self) -> Result[None, ModuleError]:
        logger.info('{path}: Module setup'.format(path=self._path))

        for component in self._components.values():
            comp_conf = self._conf.components[component.name]

            for res_type in comp_conf.resource_grants:
                if res := self._resources.get(res_type).unwrap_or_none():
                    component.add_resource(res_type, res)

            for prov_conf in comp_conf.provider_grants:
                for prov_type in prov_conf.providers:
                    for _, prov in self._providers.items(
                        lambda t, pt=prov_type: issubclass(t, pt)
                    ):
                        component.add_provider(prov_type, prov)

            if not (
                res := component.setup().map_err(
                    lambda e: ModuleError.Component(path=self._path, details=e.message)
                )
            ):
                return res

            for prov_conf in comp_conf.provider_imports:
                for prov_type in prov_conf.providers:
                    if prov := component.get_provider(prov_type).unwrap_or_none():
                        self._providers.add(prov_type, prov)

        for op in self._operators:
            if not (
                res := self._rm.setup_resource(op, self._providers).map_err(
                    lambda e: ModuleError.Resource(path=self._path, details=e.message)
                )
            ):
                return res

        logger.info('{path}: Module setup successful'.format(path=self._path))
        return Ok()

    def teardown(self) -> Result[None, ModuleError]:
        logger.info('{path}: Module teardown'.format(path=self._path))

        if not (
            res := self._rm.teardown().map_err(
                lambda e: ModuleError.Resource(path=self._path, details=e.message)
            )
        ):
            return res

        for component in reversed(self._components.values()):
            comp_conf = self._conf.components[component.name]

            for prov_conf in reversed(comp_conf.provider_imports):
                for prov_type in reversed(prov_conf.providers):
                    if prov := component.get_provider(prov_type).unwrap_or_none():
                        self._providers.remove(prov_type, prov)

            if not (
                res := component.teardown().map_err(
                    lambda e: ModuleError.Component(path=self._path, details=e.message)
                )
            ):
                return res

            for prov_conf in reversed(comp_conf.provider_grants):
                for prov_type in reversed(prov_conf.providers):
                    component.remove_provider(prov_type)

            for res_type in reversed(comp_conf.resource_grants):
                component.remove_resource(res_type)

        logger.info('{path}: Module teardown successful'.format(path=self._path))
        return Ok()



from ...logging import logger
from ...utils.imports import import_path
from ...utils.injectors import ParamsInjector
from ...utils.invoke import invoke_sync
from ..entities.application import Application


def load_packages(app: Application, key: str, **kwargs) -> None:
    from tomlkit import parse

    packages = {}
    injector = ParamsInjector()
    pyproject_toml_path = app.get_cwd() / 'pyproject.toml'

    if pyproject_toml_path.exists():
        with open(str(pyproject_toml_path)) as f:
            pyproject_toml = parse(string=f.read())

        if project := pyproject_toml.get('project'):
            if key:
                if section := project.get(key):
                    packages = section.get('packages', dict())
            else:
                packages = project.get('packages', dict())

    key = key or 'package'

    for name, dotted_path in packages.items():
        logger.info(f'Import package: "{name}"')
        dotted_path = dotted_path if ':' in dotted_path else f'{dotted_path}:{key}_init'

        if entrypoint := import_path(dotted_path):
            logger.info(f'Application "{app.name}": load package "{name}"')
            injected_args = injector(entrypoint, values={'app': app, **kwargs})
            invoke_sync(entrypoint, *injected_args.args, **injected_args.kwargs)



from abc import ABC, abstractmethod
from collections.abc import Awaitable
from typing import TypeVar, override

from greyhorse.app.abc.operators import Operator
from greyhorse.app.abc.providers import (
    FactoryProvider,
    ForwardProvider,
    MutProvider,
    Provider,
    SharedProvider,
)
from greyhorse.result import Err, Ok, Result
from greyhorse.utils.types import TypeWrapper


class ResourceMapper[T](TypeWrapper[T], ABC):
    __slots__ = ('_provider', '_operator', '_methods')

    def __init__(self, provider: Provider[T], operator: Operator[T]) -> None:
        self._provider = provider
        self._operator = operator

    def __class_getitem__(cls, provider_type: type[Provider]):
        if isinstance(provider_type, TypeVar):
            # noinspection PyUnresolvedReferences
            return super(TypeWrapper, cls).__class_getitem__(provider_type)

        class_ = super().__class_getitem__(provider_type)
        class_._methods = cls._get_provider_methods(provider_type)
        return class_

    @staticmethod
    def _get_provider_methods(provider_type: type[Provider]):
        if issubclass(provider_type, SharedProvider):
            return 'borrow', 'reclaim'
        if issubclass(provider_type, MutProvider):
            return 'acquire', 'release'
        if issubclass(provider_type, FactoryProvider):
            return 'create', 'destroy'
        if issubclass(provider_type, ForwardProvider):
            return 'take', 'drop'
        return None

    @abstractmethod
    def setup(self) -> Result[bool, str] | Awaitable[Result[bool, str]]: ...

    @abstractmethod
    def teardown(self) -> Result[None, str] | Awaitable[Result[None, str]]: ...


class SyncResourceMapper[T](ResourceMapper[T]):
    @override
    def setup(self) -> Result[None, str]:
        init_method = getattr(self._provider, self._methods[0])
        return (
            init_method()
            .map_err(lambda e: e.message)
            .map(self._operator.accept)
            .and_then(
                lambda v: Ok()
                if v
                else Err(f'Could not accept resource "{self.__wrapped_type__.__name__}"')
            )
        )

    @override
    def teardown(self) -> Result[None, str]:
        fini_method = getattr(self._provider, self._methods[1])
        return (
            self._operator.revoke()
            .map(fini_method)
            .ok_or(f'Could not revoke resource "{self.__wrapped_type__.__name__}"')
        )


class AsyncResourceMapper[T](ResourceMapper[T]):
    @override
    async def setup(self) -> Result[None, str]:
        init_method = getattr(self._provider, self._methods[0])
        return await (
            await (await init_method())
            .map_err(lambda e: e.message)
            .map_async(self._operator.accept)
        ).and_then_async(
            lambda v: Ok()
            if v
            else Err(f'Could not accept resource "{self.__wrapped_type__.__name__}"')
        )

    @override
    async def teardown(self) -> Result[None, str]:
        fini_method = getattr(self._provider, self._methods[1])
        return (await (await self._operator.revoke()).map_async(fini_method)).ok_or(
            f'Could not revoke resource "{self.__wrapped_type__.__name__}"'
        )



from types import EllipsisType, NoneType, UnionType, new_class
from typing import TypeVar, TypeVarTuple

from .strings import capitalize

_TYPES_CACHE = {}


class TypeWrapper[T]:
    __wrapped_type__: type | tuple[type, ...] = type

    @property
    def wrapped_type(self) -> type | tuple[type, ...]:
        return self.__wrapped_type__

    @classmethod
    def __generate_typename__(
        cls,
        types: type | TypeVar | TypeVarTuple | tuple[type | TypeVar | TypeVarTuple, ...],
        include_base_name: bool = True,
    ) -> str:
        if isinstance(types, tuple):
            type_args = [cls.__generate_typename__(a, False) for a in types]
            type_name = f'{''.join(type_args)}'
        elif isinstance(types, TypeVar):
            type_name = '~'
        elif isinstance(types, NoneType):
            type_name = 'None'
        elif isinstance(types, UnionType):
            type_args = [cls.__generate_typename__(a, False) for a in types.__args__]
            type_name = f'{''.join(type_args)}Union'
        elif isinstance(types, EllipsisType):
            type_name = '...'
        elif hasattr(types, '__args__'):
            type_args = [cls.__generate_typename__(a, False) for a in types.__args__]
            type_name = f'{''.join(type_args)}{capitalize(types.__name__)}'
        else:
            type_name = f'{capitalize(types.__name__)}'

        if include_base_name:
            type_name = type_name + cls.__name__
        return type_name

    def __init_subclass__(cls, **kwargs):
        if spec := kwargs.pop('spec', None):
            if not isinstance(spec, list):
                spec = [spec]

            for spec_class in spec:  # type: type
                type_name = cls.__generate_typename__(spec_class, False)
                _TYPES_CACHE[type_name] = cls

        super().__init_subclass__(**kwargs)

    def __class_getitem__(
        cls, types: type | TypeVar | TypeVarTuple | tuple[type | TypeVar | TypeVarTuple, ...]
    ):
        if isinstance(types, TypeVar | TypeVarTuple):
            # noinspection PyUnresolvedReferences
            return super().__class_getitem__(types)

        type_name = cls.__generate_typename__(types)

        if class_ := _TYPES_CACHE.get(type_name):
            return class_

        bases = [cls]

        if not isinstance(types, tuple) and hasattr(types, '__base__'):
            base_type_name = '~' + cls.__generate_typename__(types.__base__)
            if base_class := _TYPES_CACHE.get(base_type_name):
                bases = [base_class, *bases]

        if hasattr(cls, '__slots__'):
            cls.__slots__ = (*cls.__slots__, '__wrapped_type__')
            attrs = {k: v for k, v in cls.__dict__.items() if k in cls.__slots__}
            attrs['__wrapped_type__'] = types
            attrs['__module__'] = cls.__module__
            class_ = new_class(type_name, tuple(bases), exec_body=lambda d: d.update(attrs))

        else:
            attrs = dict(cls.__dict__)
            attrs.pop('__class_getitem__', None)
            attrs['__wrapped_type__'] = types
            attrs['__module__'] = cls.__module__

            class_ = new_class(type_name, tuple(bases), exec_body=lambda d: d.update(attrs))

        _TYPES_CACHE[type_name] = class_
        return class_



from copy import deepcopy
from functools import partial

from greyhorse.app.abc.providers import MutProvider, SharedProvider
from greyhorse.app.boxes import OwnerCtxRefBox
from greyhorse.app.contexts import MutCtxCallbacks, SyncContext, SyncMutContextWithCallbacks
from greyhorse.app.entities.services import SyncService, provider
from greyhorse.maybe import Just
from greyhorse.result import Ok

from ..common.resources import DictResContext, DictResource, MutDictResContext


class DictResourceBox(OwnerCtxRefBox[DictResource, DictResource]):
    allow_borrow_when_acquired = True
    allow_acq_when_borrowed = True
    allow_multiple_acquisition = False


class DictProviderService(SyncService):
    def __init__(self) -> None:
        super().__init__()
        self._value = {}
        self._box = DictResourceBox(
            SyncContext,
            SyncMutContextWithCallbacks,
            partial(deepcopy, self._value),
            partial(deepcopy, self._value),
            mut_params=dict(callbacks=MutCtxCallbacks(on_apply=Just(self._setter))),
        )

    def _setter(self, value: DictResource) -> None:
        self._value.clear()
        self._value.update(value)

    @provider(SharedProvider[DictResContext])
    def create_dict(self):
        return self._box

    @provider(MutProvider[MutDictResContext])
    def create_mut_dict(self):
        return Ok(self._box)



import inspect
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from typing import Any

from greyhorse.error import Error, ErrorCase
from greyhorse.result import Result

from .collectors import Collector, MutCollector
from .visitor import Visitor


class ControllerError(Error):
    namespace = 'greyhorse.app'

    Factory = ErrorCase(msg='Controller factory error: "{details}"', details=str)
    NoSuchResource = ErrorCase(msg='Resource "{name}" is not set', name=str)


type ControllerFactoryFn = Callable[[...], Controller | Result[Controller, ControllerError]]
type ControllerFactories = dict[type[Controller], ControllerFactoryFn]


@dataclass(slots=True, frozen=True)
class OperatorMember:
    class_name: str
    method_name: str
    resource_type: type
    method: classmethod
    params: dict[str, type] = field(default_factory=dict)
    ret_type: type | None = None


def _is_operator_member(member: classmethod) -> bool:
    return hasattr(member, '__operator__')


class Controller(ABC):
    def __init__(self) -> None:
        self._operator_members: dict[type, OperatorMember] = {}
        self._init_operator_members()

    def accept_visitor(self, visitor: Visitor) -> None:
        visitor.visit_controller(self)

    def _init_operator_members(self) -> None:
        operator_members = inspect.getmembers(self, _is_operator_member)

        for _, member in operator_members:
            member = member.__operator__
            self._operator_members[member.resource_type] = member

    def inspect(self, callback: Callable[[OperatorMember], Any]) -> None:
        for member in self._operator_members.values():
            callback(member)

    @abstractmethod
    def setup(
        self, collector: Collector[type, Any]
    ) -> Result[bool, ControllerError] | Awaitable[Result[bool, ControllerError]]: ...

    @abstractmethod
    def teardown(
        self, collector: MutCollector[type, Any]
    ) -> Result[bool, ControllerError] | Awaitable[Result[bool, ControllerError]]: ...



from collections import OrderedDict, defaultdict
from functools import partial

import networkx as nx

from ...error import Error, ErrorCase
from ...result import Err, Ok, Result
from ...utils.injectors import ParamsInjector
from ...utils.invoke import invoke_sync
from ..abc.controllers import Controller, OperatorMember
from ..abc.operators import Operator
from ..abc.providers import (
    FactoryProvider,
    ForwardProvider,
    MutProvider,
    Provider,
    SharedProvider,
)
from ..abc.selectors import Selector
from ..abc.services import ProviderMember, Service
from ..registries import MutDictRegistry
from .mappers import SyncResourceMapper


class ResourceError(Error):
    namespace = 'greyhorse.app'

    Provision = ErrorCase(msg='Resource provision error: "{details}"', details=str)
    NoSuchResource = ErrorCase(msg='No such resource: "{resource}"', resource=str)
    NoSuchDependency = ErrorCase(msg='No such dependency: "{resource}"', resource=str)


class ResourceManager:
    def __init__(self) -> None:
        self._services: list[Service] = []
        self._controllers: list[Controller] = []
        self._deps = set()
        self._resource_graph = nx.DiGraph()
        self._provided_resources = OrderedDict[Operator, SyncResourceMapper]()
        self._cached_providers = MutDictRegistry[type[Provider], Provider]()
        self._res_providers: dict[type, list[type[Provider]]] = defaultdict(list)
        self._operator_map: dict[type, dict[Controller, OperatorMember]] = defaultdict(dict)
        self._public_operators: list[Operator] = []

    def get_operators(self):
        return self._public_operators.copy()

    def add_service(self, service: Service) -> bool:
        if self._services.count(service):
            return False
        self._services.append(service)
        service.inspect(partial(self._add_service_provider, service))
        return True

    def remove_service(self, service: Service) -> bool:
        if not self._services.count(service):
            return False
        service.inspect(partial(self._remove_service_provider, service))
        self._services.remove(service)
        return True

    def add_controller(self, controller: Controller) -> bool:
        if self._controllers.count(controller):
            return False
        self._controllers.append(controller)
        controller.inspect(partial(self._add_ctrl_operator, controller))
        return True

    def remove_controller(self, controller: Controller) -> bool:
        if not self._controllers.count(controller):
            return False
        controller.inspect(partial(self._remove_ctrl_operator, controller))
        self._controllers.remove(controller)
        return True

    def setup(
        self, providers: Selector[type[Provider], Provider] | None = None
    ) -> Result[None, ResourceError]:
        operators = []

        for ctrl_dict in self._operator_map.values():
            for ctrl, operator_member in ctrl_dict.items():
                factory = getattr(ctrl, operator_member.method_name)
                match invoke_sync(factory):
                    case Ok(op) | (Operator() as op):
                        operators.append(op)

                    case Err(e):
                        return ResourceError.Provision(details=e).to_result()

                    case _:
                        continue

        for op in operators:
            if not self.setup_resource(op, providers):
                self._public_operators.append(op)

        return Ok()

    def teardown(self) -> Result[None, ResourceError]:
        while operator := next(reversed(self._provided_resources.keys()), None):
            mapper = self._provided_resources[operator]

            if not (
                res := mapper.teardown().map_err(
                    lambda e: ResourceError.Provision(details=e.message)
                )
            ):
                return res

            del self._provided_resources[operator]

        for k, _ in self._cached_providers.items():  # noqa: PERF102
            self._cached_providers.remove(k)

        self._public_operators = []
        self._deps.clear()
        self._resource_graph.clear()
        assert len(self._provided_resources) == 0
        return Ok()

    def find_provider(
        self,
        prov_type: type[Provider],
        providers: Selector[type[Provider], Provider] | None = None,
    ) -> Result[Provider, ResourceError]:
        if (prov := self._cached_providers.get(prov_type).unwrap_or_none()) or (
            prov := providers.get(prov_type).unwrap_or_none()
        ):
            return Ok(prov)

        if not self._resource_graph.has_node(prov_type):
            return ResourceError.NoSuchResource(resource=prov_type.__name__).to_result()

        node = self._resource_graph.nodes[prov_type]
        injector = ParamsInjector()

        for dep_type in self._resource_graph.successors(prov_type):
            if prov := self.find_provider(dep_type, providers):
                injector.add_type_provider(dep_type, prov.unwrap())
            else:
                return ResourceError.NoSuchDependency(resource=dep_type.__name__).to_result()

        factory = node['factory']
        injected_args = injector(factory)

        match invoke_sync(factory, *injected_args.args, **injected_args.kwargs):
            case Ok(prov) | (Provider() as prov):
                self._cached_providers.add(prov_type, prov)
                return Ok(prov)

            case Err(e):
                return ResourceError.Provision(details=e.message).to_result()

            case None:
                return ResourceError.NoSuchResource(resource=prov_type.__name__).to_result()

            case _:
                raise AssertionError()

    def setup_resource[T](
        self, operator: Operator[T], providers: Selector[type[Provider], Provider] | None = None
    ) -> Result[bool, ResourceError]:
        res_type = operator.wrapped_type
        prov_type = None

        if operator in self._provided_resources:
            return Ok(False)

        if issubclass(res_type, Provider):
            prov_type = res_type
        else:
            found = False

            if res_type in self._res_providers:
                candidates = self._res_providers[res_type]
            else:
                candidates = (
                    SharedProvider[res_type],
                    ForwardProvider[res_type],
                    FactoryProvider[res_type],
                    MutProvider[res_type],
                )

            for prov_type in candidates:
                if (
                    self._resource_graph.has_node(prov_type)
                    or self._cached_providers.has(prov_type)
                    or providers.has(prov_type)
                ):
                    found = True
                    break

            if not found:
                return ResourceError.NoSuchResource(resource=res_type.__name__).to_result()

        if (prov := self._cached_providers.get(prov_type).unwrap_or_none()) or (
            prov := providers.get(prov_type).unwrap_or_none()
        ):
            mapper = SyncResourceMapper[prov_type](prov, operator)

        else:
            if not (res := self.find_provider(prov_type, providers)):
                return res

            prov = res.unwrap()
            mapper = SyncResourceMapper[prov_type](prov, operator)

        if res := mapper.setup():
            self._provided_resources[operator] = mapper

        return res.map(lambda _: True).map_err(lambda e: ResourceError.Provision(details=e))

    def teardown_resource[T](self, operator: Operator[T]) -> Result[bool, ResourceError]:
        if not (mapper := self._provided_resources.get(operator)):
            return Ok(False)

        if res := mapper.teardown():
            del self._provided_resources[operator]

        return res.map(lambda _: True).map_err(lambda e: ResourceError.Provision(details=e))

    def _add_service_provider(self, service: Service, provider_member: ProviderMember) -> None:
        if self._resource_graph.has_node(provider_member.provider_type):
            return

        self._res_providers[provider_member.resource_type].append(provider_member.provider_type)

        self._resource_graph.add_node(
            provider_member.provider_type, factory=getattr(service, provider_member.method_name)
        )

        for param_type in provider_member.params.values():
            if not self._resource_graph.has_node(param_type):
                self._deps.add(param_type)

            self._resource_graph.add_edge(provider_member.provider_type, param_type)

    def _remove_service_provider(self, _: Service, provider_member: ProviderMember) -> None:
        if not self._resource_graph.has_node(provider_member.provider_type):
            return

        self._res_providers[provider_member.resource_type].remove(provider_member.provider_type)
        self._resource_graph.remove_node(provider_member.provider_type)

        for param_type in provider_member.params.values():
            if not self._resource_graph.has_node(param_type):
                self._deps.remove(param_type)

    def _add_ctrl_operator(
        self, controller: Controller, operator_member: OperatorMember
    ) -> None:
        ctrl_dict = self._operator_map[operator_member.resource_type]

        if controller in ctrl_dict:
            return

        ctrl_dict[controller] = operator_member

    def _remove_ctrl_operator(
        self, controller: Controller, operator_member: OperatorMember
    ) -> None:
        ctrl_dict = self._operator_map[operator_member.resource_type]

        if controller not in ctrl_dict:
            return

        del ctrl_dict[controller]



import pytest
from greyhorse.enum import Enum, Struct, Tuple, Unit


class Result[T, E](Enum):
    Ok = Tuple(T)
    Err = Tuple(E)
    Val = Struct(msg='message', value=int)
    Uno = Unit()


class ResultAllowInit[T, E](Enum, allow_init=True):
    pass


def test_enum() -> None:
    with pytest.raises(NotImplementedError):
        Result()

    ResultAllowInit()

    r1 = Result.Ok(1)
    r2 = Result.Err('1')
    r3 = Result.Val(value=123)
    r4 = Result.Uno

    res = 0

    for r in (r1, r2, r3, r4):
        match r:
            case Result.Ok(a):
                assert a == 1
                res += 1
            case Result.Err(b):
                assert b == '1'
                res += 1
            case Result.Val(c):
                assert c == 123
                res += 1
            case Result.Uno:
                res += 1

    assert res == 4



from abc import ABC, abstractmethod
from typing import Self


class Query(ABC):
    @abstractmethod
    def apply_filter[F](self, filter: F) -> F: ...

    @abstractmethod
    def apply_sorting[S](self, sorting: S) -> S: ...

    @abstractmethod
    def __eq__(self, other: Self): ...

    def __ne__(self, other: Self):
        return not (self == other)



from unittest import mock

import pytest
from greyhorse.error import Error, ErrorCase
from greyhorse.maybe import Just, Nothing
from greyhorse.result import (
    Err,
    Ok,
    Result,
    ResultUnwrapError,
    as_result_async,
    as_result_sync,
    do,
    do_async,
)


class TestError(Error):
    Unexpected = ErrorCase(msg='Unexpected error')


def test_result() -> None:
    result_ok = Result(123)
    result_err = Result(TestError.Unexpected())

    assert result_ok
    assert not result_err

    assert result_ok.is_ok()
    assert not result_err.is_ok()

    assert not result_ok.is_err()
    assert result_err.is_err()

    result_ok = Ok(123)
    result_err = Err('err')

    assert result_ok
    assert not result_err

    assert result_ok.is_ok()
    assert not result_err.is_ok()

    assert result_ok.is_ok_and(lambda v: v == 123)
    assert not result_ok.is_ok_and(lambda v: v != 123)
    assert not result_err.is_ok_and(lambda v: v == 123)
    assert not result_err.is_ok_and(lambda v: v != 123)

    assert not result_ok.is_err()
    assert result_err.is_err()

    assert not result_ok.is_err_and(lambda e: e == 'err')
    assert not result_ok.is_err_and(lambda e: e != 'err')
    assert result_err.is_err_and(lambda e: e == 'err')
    assert not result_err.is_err_and(lambda e: e != 'err')

    assert Just(123) == result_ok.ok()
    assert Nothing is result_err.ok()

    assert Nothing is result_ok.err()
    assert Just('err') == result_err.err()

    inspect_mock = mock.Mock()

    assert result_ok is result_ok.inspect(inspect_mock)
    assert result_err is result_err.inspect(inspect_mock)

    inspect_mock.assert_called_once_with(123)

    inspect_mock.reset_mock()

    assert result_ok is result_ok.inspect_err(inspect_mock)
    assert result_err is result_err.inspect_err(inspect_mock)

    inspect_mock.assert_called_once_with('err')

    assert result_ok.expect('exception') == 123

    with pytest.raises(ResultUnwrapError) as excinfo:
        result_err.expect('exception')

    assert str(excinfo.value) == 'exception'
    assert excinfo.value.value == 'err'

    with pytest.raises(ResultUnwrapError) as excinfo:
        result_ok.expect_err('exception')

    assert str(excinfo.value) == 'exception'
    assert excinfo.value.value == 123

    assert result_err.expect_err('exception') == 'err'

    assert result_ok.unwrap() == 123

    with pytest.raises(ResultUnwrapError):
        result_err.unwrap()

    with pytest.raises(ResultUnwrapError):
        result_ok.unwrap_err()

    assert result_err.unwrap_err() == 'err'

    assert result_ok.unwrap_or(456) == 123
    assert result_err.unwrap_or(456) == 456

    assert result_ok.unwrap_or_none() == 123
    assert None is result_err.unwrap_or_none()

    assert None is result_ok.unwrap_err_or_none()
    assert result_err.unwrap_err_or_none() == 'err'

    assert result_ok.unwrap_or_else(lambda e: 789) == 123
    assert result_err.unwrap_or_else(lambda e: 789) == 789

    assert result_ok.unwrap_or_raise(Exception) == 123

    with pytest.raises(Exception):
        result_err.unwrap_or_raise(Exception)

    assert Ok(134) == result_ok.map(lambda i: i + 11)
    assert Err('err') == result_err.map(lambda i: i + 11)

    assert result_ok.map_or(234, lambda i: i + 1) == 124
    assert result_err.map_or(234, lambda i: i + 1) == 234

    assert result_ok.map_or_else(lambda e: 235, lambda i: i + 2) == 125
    assert result_err.map_or_else(lambda e: 235, lambda i: i + 2) == 235

    assert Ok(123) == result_ok.map_err(lambda e: e.upper())
    assert Err('ERR') == result_err.map_err(lambda e: e.upper())

    assert result_ok is result_ok.and_(result_ok)
    assert result_err is result_ok.and_(result_err)
    assert result_err is result_err.and_(result_ok)
    assert result_err is result_err.and_(result_err)

    assert Ok(456) == result_ok.and_then(lambda v: Ok(456))
    assert result_err is result_err.and_then(lambda v: Ok(456))

    assert result_ok is result_ok.or_(result_ok)
    assert result_ok is result_ok.or_(result_err)
    assert result_ok is result_err.or_(result_ok)
    assert result_err is result_err.or_(result_err)

    assert result_ok is result_ok.or_else(lambda e: Ok(456))
    assert Ok(456) == result_err.or_else(lambda e: Ok(456))

    assert result_ok is result_ok.flatten()
    assert result_ok is Ok(result_ok).flatten()
    assert result_err is result_err.flatten()
    assert result_err is Ok(result_err).flatten()

    assert Just(result_ok) == result_ok.to_maybe()
    assert Just(Ok(123)) == Ok(Just(123)).to_maybe()
    assert Just(Err('err')) == Err('err').to_maybe()
    assert Nothing is Ok(Nothing).to_maybe()


def test_as_result_sync() -> None:
    @as_result_sync(ValueError, IndexError)
    def f(value: int) -> int:
        if value == 0:
            raise ValueError  # becomes Err
        if value == 1:
            raise IndexError  # becomes Err
        if value == 2:
            raise KeyError  # raises Exception
        return value  # becomes Ok

    assert isinstance(f(0).unwrap_err(), ValueError)
    assert isinstance(f(1).unwrap_err(), IndexError)

    with pytest.raises(Exception):
        f(2)

    assert Ok(3) == f(3)


@pytest.mark.asyncio
async def test_as_result_async() -> None:
    @as_result_async(ValueError, IndexError)
    async def f(value: int) -> int:
        if value == 0:
            raise ValueError  # becomes Err
        if value == 1:
            raise IndexError  # becomes Err
        if value == 2:
            raise KeyError  # raises Exception
        return value  # becomes Ok

    assert isinstance((await f(0)).unwrap_err(), ValueError)
    assert isinstance((await f(1)).unwrap_err(), IndexError)

    with pytest.raises(Exception):
        await f(2)

    assert Ok(3) == await f(3)


def test_do_sync() -> None:
    def get_result(i: int = 1):
        if i < 3:
            return Ok(i)
        return Err(i)

    final_result: Result[int, str] = do(Ok(x + y) for x in get_result(1) for y in get_result(2))

    assert Ok(3) == final_result

    final_result: Result[int, str] = do(
        Ok(x + y + z) for x in get_result(1) for y in get_result(2) for z in get_result(10)
    )

    assert Err(10) == final_result


@pytest.mark.asyncio
async def test_do_async() -> None:
    async def get_result(i: int = 1):
        if i < 3:
            return Ok(i)
        return Err(i)

    final_result: Result[int, str] = await do_async(
        Ok(x + y) for x in await get_result(1) for y in await get_result(2)
    )

    assert Ok(3) == final_result

    final_result: Result[int, str] = await do_async(
        Ok(x + y + z)
        for x in await get_result(1)
        for y in await get_result(2)
        for z in await get_result(10)
    )

    assert Err(10) == final_result



from greyhorse.i18n import StaticTranslator


def test_load_file_unload_file() -> None:
    tr = StaticTranslator()
    assert tr.size() == 0

    tr.load_file('tests/translations.toml')
    assert tr.size() == 8
    tr.load_file('tests/translations.toml', namespace='root')
    assert tr.size() == 16

    assert tr('translations.title') == 'Title'
    assert tr('root.title') == 'Title'

    assert not tr.set_default_lang('translations1', 'ru')
    assert tr.set_default_lang('translations', 'ru')

    assert tr('translations.title') == 'Заголовок'
    assert tr('root.title') == 'Title'

    tr.unload('translations1')
    tr.unload('translations')

    assert tr.size() == 8
    assert not tr('translations.title')
    assert tr('root.title') == 'Title'


def test_one_namespace() -> None:
    tr = StaticTranslator()
    tr.load_file('tests/translations.toml')

    assert not tr('default')
    assert tr('title') == 'Title'
    assert tr('translations.title') == 'Title'
    assert tr('title', 'en') == 'Title'
    assert tr('title', 'ru') == 'Заголовок'
    assert tr('translations.title') == 'Title'
    assert tr('translations.title', 'en') == 'Title'
    assert tr('translations.title', 'ru') == 'Заголовок'

    assert tr('owner.name') == 'Name'
    assert tr('owner.name', 'en') == 'Name'
    assert tr('owner.name', 'ru') == 'Имя'
    assert tr('translations.owner.name') == 'Name'
    assert tr('translations.owner.name', 'en') == 'Name'
    assert tr('translations.owner.name', 'ru') == 'Имя'

    assert tr('owner.name.en') == 'Name'
    assert tr('owner.name.ru') == 'Имя'
    assert tr('translations.owner.name.en') == 'Name'
    assert tr('translations.owner.name.ru') == 'Имя'

    assert tr('owner.name.en', 'en') == 'Name'
    assert tr('owner.name.ru', 'en') == 'Имя'
    assert tr('owner.name.en', 'ru') == 'Name'
    assert tr('owner.name.ru', 'ru') == 'Имя'
    assert tr('translations.owner.name.en', 'en') == 'Name'
    assert tr('translations.owner.name.ru', 'en') == 'Имя'
    assert tr('translations.owner.name.en', 'ru') == 'Name'
    assert tr('translations.owner.name.ru', 'ru') == 'Имя'

    assert tr('data.title1') == 'Data'
    assert tr('data.title1', 'en') == 'Data'
    assert tr('data.title1', 'ru') == 'Data'
    assert tr('translations.data.title1') == 'Data'
    assert tr('translations.data.title1', 'en') == 'Data'
    assert tr('translations.data.title1', 'ru') == 'Data'

    assert not tr('data.title1.en')
    assert not tr('data.title1.ru')
    assert not tr('data.title1.en', 'en')
    assert not tr('data.title1.en', 'ru')
    assert not tr('data.title1.ru', 'en')
    assert not tr('data.title1.ru', 'ru')
    assert not tr('translations.data.title1.en')
    assert not tr('translations.data.title1.ru')
    assert not tr('translations.data.title1.en', 'en')
    assert not tr('translations.data.title1.en', 'ru')
    assert not tr('translations.data.title1.ru', 'en')
    assert not tr('translations.data.title1.ru', 'ru')

    assert not tr('data.title2')
    assert not tr('data.title2', 'en')
    assert tr('data.title2', 'ru') == 'Данные'
    assert not tr('translations.data.title2')
    assert not tr('translations.data.title2', 'en')
    assert tr('translations.data.title2', 'ru') == 'Данные'

    assert not tr('data.title2.en')
    assert tr('data.title2.ru') == 'Данные'
    assert not tr('data.title2.en', 'en')
    assert not tr('data.title2.en', 'ru')
    assert tr('data.title2.ru', 'en') == 'Данные'
    assert tr('data.title2.ru', 'ru') == 'Данные'
    assert not tr('translations.data.title2.en')
    assert tr('translations.data.title2.ru') == 'Данные'
    assert not tr('translations.data.title2.en', 'en')
    assert not tr('translations.data.title2.en', 'ru')
    assert tr('translations.data.title2.ru', 'en') == 'Данные'
    assert tr('translations.data.title2.ru', 'ru') == 'Данные'


def test_two_namespaces() -> None:
    tr = StaticTranslator()
    tr.load_file('tests/translations.toml')
    tr.load_file('tests/translations.toml', namespace='root')

    assert not tr('default')
    assert not tr('root.default')
    assert not tr('title')
    assert tr('translations.title') == 'Title'
    assert tr('root.title') == 'Title'
    assert not tr('title', 'en')
    assert not tr('title', 'ru')
    assert tr('translations.title', 'en') == 'Title'
    assert tr('translations.title', 'ru') == 'Заголовок'
    assert tr('root.title', 'en') == 'Title'
    assert tr('root.title', 'ru') == 'Заголовок'

    assert not tr('owner.name')
    assert not tr('owner.name', 'en')
    assert not tr('owner.name', 'ru')
    assert tr('translations.owner.name') == 'Name'
    assert tr('translations.owner.name', 'en') == 'Name'
    assert tr('translations.owner.name', 'ru') == 'Имя'
    assert tr('root.owner.name') == 'Name'
    assert tr('root.owner.name', 'en') == 'Name'
    assert tr('root.owner.name', 'ru') == 'Имя'

    assert not tr('owner.name.en')
    assert not tr('owner.name.ru')
    assert tr('translations.owner.name.en') == 'Name'
    assert tr('translations.owner.name.ru') == 'Имя'
    assert tr('root.owner.name.en') == 'Name'
    assert tr('root.owner.name.ru') == 'Имя'

    assert not tr('owner.name.en', 'en')
    assert not tr('owner.name.ru', 'en')
    assert not tr('owner.name.en', 'ru')
    assert not tr('owner.name.ru', 'ru')
    assert tr('translations.owner.name.en', 'en') == 'Name'
    assert tr('translations.owner.name.ru', 'en') == 'Имя'
    assert tr('translations.owner.name.en', 'ru') == 'Name'
    assert tr('translations.owner.name.ru', 'ru') == 'Имя'
    assert tr('root.owner.name.en', 'en') == 'Name'
    assert tr('root.owner.name.ru', 'en') == 'Имя'
    assert tr('root.owner.name.en', 'ru') == 'Name'
    assert tr('root.owner.name.ru', 'ru') == 'Имя'

    assert not tr('data.title1')
    assert not tr('data.title1', 'en')
    assert not tr('data.title1', 'ru')
    assert tr('translations.data.title1') == 'Data'
    assert tr('translations.data.title1', 'en') == 'Data'
    assert tr('translations.data.title1', 'ru') == 'Data'
    assert tr('root.data.title1') == 'Data'
    assert tr('root.data.title1', 'en') == 'Data'
    assert tr('root.data.title1', 'ru') == 'Data'

    assert not tr('data.title1.en')
    assert not tr('data.title1.ru')
    assert not tr('data.title1.en', 'en')
    assert not tr('data.title1.en', 'ru')
    assert not tr('data.title1.ru', 'en')
    assert not tr('data.title1.ru', 'ru')
    assert not tr('translations.data.title1.en')
    assert not tr('translations.data.title1.ru')
    assert not tr('translations.data.title1.en', 'en')
    assert not tr('translations.data.title1.en', 'ru')
    assert not tr('translations.data.title1.ru', 'en')
    assert not tr('translations.data.title1.ru', 'ru')
    assert not tr('root.data.title1.en')
    assert not tr('root.data.title1.ru')
    assert not tr('root.data.title1.en', 'en')
    assert not tr('root.data.title1.en', 'ru')
    assert not tr('root.data.title1.ru', 'en')
    assert not tr('root.data.title1.ru', 'ru')

    assert not tr('data.title2')
    assert not tr('data.title2', 'en')
    assert not tr('data.title2', 'ru')
    assert not tr('translations.data.title2')
    assert not tr('translations.data.title2', 'en')
    assert tr('translations.data.title2', 'ru') == 'Данные'
    assert not tr('root.data.title2')
    assert not tr('root.data.title2', 'en')
    assert tr('root.data.title2', 'ru') == 'Данные'

    assert not tr('data.title2.en')
    assert not tr('data.title2.ru')
    assert not tr('data.title2.en', 'en')
    assert not tr('data.title2.en', 'ru')
    assert not tr('data.title2.ru', 'en')
    assert not tr('data.title2.ru', 'ru')
    assert not tr('translations.data.title2.en')
    assert tr('translations.data.title2.ru') == 'Данные'
    assert not tr('translations.data.title2.en', 'en')
    assert not tr('translations.data.title2.en', 'ru')
    assert tr('translations.data.title2.ru', 'en') == 'Данные'
    assert tr('translations.data.title2.ru', 'ru') == 'Данные'
    assert not tr('root.data.title2.en')
    assert tr('root.data.title2.ru') == 'Данные'
    assert not tr('root.data.title2.en', 'en')
    assert not tr('root.data.title2.en', 'ru')
    assert tr('root.data.title2.ru', 'en') == 'Данные'
    assert tr('root.data.title2.ru', 'ru') == 'Данные'



from greyhorse.error import Error, ErrorCase
from greyhorse.i18n import StaticTranslator
from greyhorse.result import Result

tr = StaticTranslator()


class TestError(Error, tr=tr):
    namespace = 'tests'

    FirstError = ErrorCase(msg='First Error: "{value}"', value=int)
    SecondError = ErrorCase(msg='Second Error: "{value}"', value=int)
    TranslatedError = ErrorCase(code='owner.error', value=str)
    Unexpected = ErrorCase(msg='Unexpected error: "{details}"', details=str)


def test_error() -> None:
    e1 = TestError.FirstError(value=123)
    e2 = TestError.SecondError(value='qwer')
    e3 = TestError.TranslatedError(value='error value')
    e4 = TestError.Unexpected(details='Unexpected')

    tr.load_file('tests/translations.toml', namespace='tests')

    assert e1.message == 'First Error: "123"'
    assert e2.message == 'Second Error: "qwer"'
    assert e3.message == "Error: 'error value'"
    assert e4.message == 'Unexpected error: "Unexpected"'

    tr.unload('tests')

    res = 0

    for e in (e1, e2, e3, e4):
        match e:
            case TestError.FirstError(a):
                assert a == 123
                res += 1
            case TestError.SecondError(b):
                assert b == 'qwer'
                res += 1
            case TestError.TranslatedError(c):
                assert c == 'error value'
                res += 1
            case TestError.Unexpected(d):
                assert d == 'Unexpected'
                res += 1

        err_result = e.to_result()
        assert isinstance(err_result, Result.Err)
        assert err_result.is_err()

    assert res == 4



from unittest import mock

import pytest
from greyhorse.maybe import Just, Maybe, MaybeUnwrapError, Nothing
from greyhorse.result import Err, Ok


def test_maybe() -> None:
    maybe_just = Maybe(123)
    maybe_none = Maybe(None)

    assert maybe_just
    assert not maybe_none

    assert maybe_just.is_just()
    assert not maybe_none.is_just()

    assert not maybe_just.is_nothing()
    assert maybe_none.is_nothing()

    maybe_just = Just(123)
    maybe_none = Nothing

    assert maybe_just
    assert not maybe_none

    assert maybe_just.is_just()
    assert not maybe_none.is_just()

    assert maybe_just.is_just_and(lambda v: v == 123)
    assert not maybe_just.is_just_and(lambda v: v == 456)
    assert not maybe_none.is_just_and(lambda v: v == 123)
    assert not maybe_none.is_just_and(lambda v: v == 456)

    assert not maybe_just.is_nothing()
    assert maybe_none.is_nothing()

    assert Ok(123) == maybe_just.ok_or('err')
    assert Err('err') == maybe_none.ok_or('err')

    assert Ok(123) == maybe_just.ok_or_else(lambda: 'err')
    assert Err('err') == maybe_none.ok_or_else(lambda: 'err')

    inspect_mock = mock.Mock()

    assert maybe_just is maybe_just.inspect(inspect_mock)
    assert maybe_none is maybe_none.inspect(inspect_mock)

    inspect_mock.assert_called_once_with(123)

    assert maybe_just.expect('exception') == 123

    with pytest.raises(MaybeUnwrapError) as excinfo:
        maybe_none.expect('exception')

    assert str(excinfo.value) == 'exception'

    assert maybe_just.unwrap() == 123

    with pytest.raises(MaybeUnwrapError):
        maybe_none.unwrap()

    assert maybe_just.unwrap_or(456) == 123
    assert maybe_none.unwrap_or(456) == 456

    assert maybe_just.unwrap_or_none() == 123
    assert None is maybe_none.unwrap_or_none()

    assert maybe_just.unwrap_or_else(lambda: 789) == 123
    assert maybe_none.unwrap_or_else(lambda: 789) == 789

    assert maybe_just.unwrap_or_raise(Exception) == 123

    with pytest.raises(Exception):
        maybe_none.unwrap_or_raise(Exception)

    assert Just(134) == maybe_just.map(lambda i: i + 11)
    assert Nothing is maybe_none.map(lambda i: i + 11)

    assert maybe_just.map_or(234, lambda i: i + 1) == 124
    assert maybe_none.map_or(234, lambda i: i + 1) == 234

    assert maybe_just.map_or_else(lambda: 235, lambda i: i + 2) == 125
    assert maybe_none.map_or_else(lambda: 235, lambda i: i + 2) == 235

    assert maybe_just is maybe_just.and_(maybe_just)
    assert maybe_none is maybe_just.and_(maybe_none)
    assert maybe_none is maybe_none.and_(maybe_just)
    assert maybe_none is maybe_none.and_(maybe_none)

    assert Just(456) == maybe_just.and_then(lambda v: Just(456))
    assert Nothing is maybe_none.and_then(lambda v: Just(456))

    assert maybe_just is maybe_just.filter(lambda v: v == 123)
    assert Nothing is maybe_just.filter(lambda v: v != 123)
    assert Nothing is maybe_none.filter(lambda v: v == 123)

    assert maybe_just is maybe_just.or_(maybe_just)
    assert maybe_just is maybe_just.or_(maybe_none)
    assert maybe_just is maybe_none.or_(maybe_just)
    assert maybe_none is maybe_none.or_(maybe_none)

    assert maybe_just is maybe_just.or_else(lambda: Just(456))
    assert Just(456) == maybe_none.or_else(lambda: Just(456))

    assert Nothing is maybe_just.xor(maybe_just)
    assert maybe_just is maybe_just.xor(maybe_none)
    assert maybe_just is maybe_none.xor(maybe_just)
    assert Nothing is maybe_none.xor(maybe_none)

    assert maybe_just is maybe_just.flatten()
    assert maybe_just is Just(maybe_just).flatten()
    assert maybe_none is maybe_none.flatten()
    assert maybe_none is Just(maybe_none).flatten()

    assert Ok(maybe_just) == maybe_just.to_result()
    assert Ok(Just(123)) == Just(Ok(123)).to_result()
    assert Err('err') == Just(Err('err')).to_result()
    assert Ok(Nothing) == Nothing.to_result()



from abc import ABC, abstractmethod
from collections.abc import Awaitable
from typing import Union

from greyhorse.error import Error, ErrorCase
from greyhorse.result import Result
from greyhorse.utils.types import TypeWrapper


class BorrowError(Error):
    namespace = 'greyhorse.app'

    Empty = ErrorCase(
        msg='Cannot borrow "{name}" as immutable because the value is empty', name=str
    )

    MovedOut = ErrorCase(
        msg='Cannot borrow "{name}" as immutable because the value was moved out', name=str
    )

    BorrowedAsMutable = ErrorCase(
        msg='Cannot borrow "{name}" as immutable because it is also borrowed as mutable',
        name=str,
    )


class BorrowMutError(Error):
    namespace = 'greyhorse.app'

    Empty = ErrorCase(
        msg='Cannot borrow "{name}" as mutable because the value is empty', name=str
    )

    MovedOut = ErrorCase(
        msg='Cannot borrow "{name}" as mutable because the value was moved out', name=str
    )

    AlreadyBorrowed = ErrorCase(
        msg='Cannot borrow "{name}" as mutable more than once at a time', name=str
    )

    BorrowedAsImmutable = ErrorCase(
        msg='Cannot borrow "{name}" as mutable because it is also borrowed as immutable',
        name=str,
    )


class FactoryError(Error):
    namespace = 'greyhorse.app'

    Internal = ErrorCase(
        msg='Cannot construct "{name}" because an internal error occurred: "{details}"',
        name=str,
        details=str,
    )


class ForwardError(Error):
    namespace = 'greyhorse.app'

    Empty = ErrorCase(msg='Cannot forward "{name}" because the value is empty', name=str)

    MovedOut = ErrorCase(
        msg='Cannot forward "{name}" because the value was moved out', name=str
    )


class Provider[T](TypeWrapper[T]): ...


class SharedProvider[T](Provider[T], ABC):
    @abstractmethod
    def borrow(self) -> Result[T, BorrowError] | Awaitable[Result[T, BorrowError]]: ...

    @abstractmethod
    def reclaim(self, instance: T) -> None | Awaitable[None]: ...


class MutProvider[T](Provider[T], ABC):
    @abstractmethod
    def acquire(self) -> Result[T, BorrowMutError] | Awaitable[Result[T, BorrowMutError]]: ...

    @abstractmethod
    def release(self, instance: T) -> None | Awaitable[None]: ...


class FactoryProvider[T](Provider[T], ABC):
    @abstractmethod
    def create(self) -> Result[T, FactoryError] | Awaitable[Result[T, FactoryError]]: ...

    @abstractmethod
    def destroy(self, instance: T) -> None | Awaitable[None]: ...


class ForwardProvider[T](Provider[T], ABC):
    @abstractmethod
    def take(self) -> Result[T, ForwardError] | Awaitable[Result[T, ForwardError]]: ...

    @abstractmethod
    def drop(self, instance: T) -> None | Awaitable[None]: ...

    @abstractmethod
    def __bool__(self) -> bool | Awaitable[bool]: ...


AnyProvider = Union[SharedProvider, MutProvider, FactoryProvider, ForwardProvider]



import hashlib
from typing import Any

from orjson import orjson

from greyhorse.utils.dicts import dict_values_to_str


def calculate_digest(data: Any, size: int = 0) -> str:
    if isinstance(data, dict):
        dumped = orjson.dumps(dict_values_to_str(data))
    else:
        dumped = str(data).encode('utf-8')

    size = size if 0 < size <= hashlib.blake2b.MAX_DIGEST_SIZE else 4 + int(len(dumped) / 32)
    digest_size = min(hashlib.blake2b.MAX_DIGEST_SIZE, size)
    hash_sum = hashlib.blake2b(dumped, digest_size=digest_size)
    return hash_sum.hexdigest()
